import { IsAny, IsNever } from 'type-fest';

type NonEmptyArray<T> = [T, ...Array<T>];
type Mapped<T extends IterableContainer, K> = {
    -readonly [P in keyof T]: K;
};
/**
 * This should only be used for defining generics which extend any kind of JS
 * array under the hood, this includes arrays *AND* tuples (of the form [x, y],
 * and of the form [x, ...y[]], etc...), and their readonly equivalent. This
 * allows us to be more inclusive to what functions can process.
 *
 * @example map<T extends ArrayLike>(items: T) { ... }
 *
 * We would've named this `ArrayLike`, but that's already used by typescript...
 * @see This was inspired by the type-definition of Promise.all (https://github.com/microsoft/TypeScript/blob/1df5717b120cddd325deab8b0f2b2c3eecaf2b01/src/lib/es2015.promise.d.ts#L21)
 */
type IterableContainer<T = unknown> = ReadonlyArray<T> | readonly [];
/**
 * We define a Simple record as one that doesn't define it's properties beyond
 * their general type. This is the case for string and number keys (and their
 * combination). Another way of thinking about simple records is that the number
 * of properties are unbound (they can be empty, or they can have 1000 props).
 *
 * TODO: Template string literals (e.g. `prefix_${number}`) should also be
 * considered simple, but we don't know how to check for them...
 */
type IfSimpleRecord<T, TypeIfSimpleRecord = true, TypeIfNotSimpleRecord = false> = string extends keyof T ? TypeIfSimpleRecord : number extends keyof T ? TypeIfSimpleRecord : number | string extends keyof T ? TypeIfSimpleRecord : TypeIfNotSimpleRecord;
/**
 * A union of all keys of T which are not symbols, and where number keys are
 * converted to strings, following the definition of `Object.keys` and
 * `Object.entries`.
 *
 * Inspired and largely copied from [`sindresorhus/ts-extras`](https://github.com/sindresorhus/ts-extras/blob/44f57392c5f027268330771996c4fdf9260b22d6/source/object-keys.ts).
 *
 * @see EnumerableStringKeyedValueOf
 */
type EnumerableStringKeyOf<T> = `${Exclude<keyof T, symbol>}`;
/**
 * A union of all values of properties in T which are not keyed by a symbol,
 * following the definition of `Object.values` and `Object.entries`.
 */
type EnumerableStringKeyedValueOf<T> = {
    [K in keyof T]-?: K extends symbol ? never : T[K];
}[keyof T];
/**
 * This is the type you'd get from doing:
 * `Object.fromEntries(Object.entries(x))`.
 */
type ReconstructedRecord<T> = Record<EnumerableStringKeyOf<T>, EnumerableStringKeyedValueOf<T>>;
/**
 * An extension of Extract for type predicates which falls back to the base
 * in order to narrow the `unknown` case.
 *
 * @example
 *   function isMyType<T>(data: T | MyType): data is NarrowedTo<T, MyType> { ... }
 */
type NarrowedTo<T, Base> = Extract<T, Base> extends never ? Base : IsAny<T> extends true ? Base : Extract<T, Base>;
type ExactRecord<Key extends PropertyKey, Value> = IfSimpleRecord<Record<Key, Value>, Record<Key, Value>, Partial<Record<Key, Value>>>;
type ReorderedArray<T extends IterableContainer> = {
    -readonly [P in keyof T]: T[number];
};
type IsUnion<T> = InternalIsUnion<T>;
type InternalIsUnion<T, U = T> = (IsNever<T> extends true ? false : T extends any ? [U] extends [T] ? false : true : never) extends infer Result ? boolean extends Result ? true : Result : never;

export type { EnumerableStringKeyedValueOf as E, IterableContainer as I, Mapped as M, NonEmptyArray as N, ReconstructedRecord as R, EnumerableStringKeyOf as a, ExactRecord as b, NarrowedTo as c, IsUnion as d, IfSimpleRecord as e, ReorderedArray as f };
